/*
 * MIT License
 *
 * Copyright (c) 2023 Murray R. Van Luyn. green@bug-eyed.monster
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

/*!
 * \file C8_CO2_5K.cpp
 * \brief Shenzhen Shenchen Technology C8-CO2-5K 
 *           CO2 sensor library for Arduino
 * \details
 *      Source:         
 *      Documentation:  
 */

#include "C8_CO2_5K.h"

/*!
 * \brief Constructor with serial Stream
 * \param serial
 *      Serial Stream pointer.
 */
C8_CO2_5K::C8_CO2_5K(Stream *serial) : _serial(serial)
{
}

/*!
 * \brief Destructor
 * \details
 *      The serial Stream pointer is cleared and requires a new constructor to reuse it again.
 */
C8_CO2_5K::~C8_CO2_5K()
{
    _serial = nullptr;
}

/*!
 * \brief Get CO2 ppm concentration from sensor
 * \retval 0
 *      C8_CO2_5K response error codes.
 * \retval 0..399 ppm
 *      Incorrect values. Minimum value starts at 400ppm outdoor fresh air.
 * \retval 400..1000 ppm
 *      Concentrations typical of occupied indoor spaces with good air exchange.
 * \retval 1000..2000 ppm
 *      Complaints of drowsiness and poor air quality. Ventilation is required.
 * \retval 2000..5000 ppm
 *      Headaches, sleepiness and stagnant, stale, stuffy air. Poor concentration, loss of
 *      attention, increased heart rate and slight nausea may also be present.\n
 *      Higher values are extremely dangerous and cannot be measured.
 */
uint16_t C8_CO2_5K::getCO2(void)
{
  int timeout = 0;
  uint16_t ppm_value = 0;
  uint16_t crc = 0;
  uint8_t readCO2[5] = {0x64, 0X69, 0X03, 0X5e, 0X4e};  
  uint8_t response[14] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};

  // Flush RX buffer.
  while(_serial->available()) _serial->read();

  // Send one request.
  _serial->write(readCO2, 5);
  
  // Wait for a 14 byte response.
  while (_serial->available() < 14) 
  {
    // If no response after ~500ms, timeout.
    timeout++;

    if (timeout > 10)
    {
      // Flush RX buffer
      while(_serial->available()) _serial->read();
      
      // No response before timeout, so exit.
      return(0); 
    }

    delay(50);
  }
  
  // Read serial RX buffer.
  for (int i = 0; i < 14; i++)
  {
    response[i] = _serial->read();
  } 

  // Serial.print("Response from sensor: ");
  // for (int i = 0; i < 14; i++)
  // {
  //   Serial.print(response[i], HEX);
  //   Serial.print(" ");
  // } 
  // Serial.println();
  
  // Get ppm data.
  ppm_value = (response[5] * 256) + response[4];

  crc = calcuCRC(response, 12);

  // Serial.print("CRC: ");
  // Serial.print(crc % 256, HEX);
  // Serial.print(" ");
  // Serial.print(crc / 256, HEX);
  // Serial.println();

  // Invalid response?
  if (!((response[0] == 0x64) &&
        (response[1] == 0x69) &&
        (response[2] == 0x03) &&
        (response[3] == 0x01) &&
       (response[12] == crc % 256) &&
       (response[13] == crc / 256)))
  {
    return(0);
  }

  // Serial.print("ppm: ");
  // Serial.print(ppm_value);
  // Serial.println();

  return(ppm_value);
}


/*!
 * \brief Single Point Calibration manually at known ppm
 * \details
 *      The sensor must be powered-up for at least 10 minutes in known concentration air at room
 *      temperature. Then call this function once to execute single point calibration.\n
* \param ppm_value
 *      Known ppm to calibrate sensor to. 
 * \retval 0
 *      Failure.
 * \retval 1
 *      Success.
 */
bool C8_CO2_5K::setCO2(uint16_t ppm_value)
{
  int timeout = 0;
  uint8_t writeCO2[6] = {0x11, 0X03, 0X03, 0X00, 0X00, 0x00};  
  uint8_t response[4] = {0, 0, 0, 0};

  // Check ppm input parameter.
  if ((ppm_value > 1500) || (ppm_value < 400)) return (false);

  // Add concentration to calibration command.
  writeCO2[3] = ppm_value / 256;
  writeCO2[4] = ppm_value % 256;

  // Add checksum to calibration command.
  writeCO2[5] = 256;
  for(int i = 0; i < 5; i++) writeCO2[5] -= writeCO2[i]; 

  // Serial.print("Command for sensor: ");
  // for (int i = 0; i < 6; i++)
  // {
  //   Serial.print(writeCO2[i], HEX);
  //   Serial.print(" ");
  // } 
  // Serial.println();

  // Flush RX buffer.
  while(_serial->available()) _serial->read();

  // Send one request.
  _serial->write(writeCO2, 6);

  // Wait for a 4 byte response.
  while (_serial->available() < 4) 
  {
    // If no response after ~500ms, timeout.
    timeout++;

    if (timeout > 20)
    {
      // Flush RX buffer
      while(_serial->available()) _serial->read();
      
      // No response before timeout, so exit.
      return(false); 
    }

    delay(50);
  }
  
  // Read serial RX buffer.
  for (int i = 0; i < 4; i++)
  {
    response[i] = _serial->read();
  } 

  // Serial.print("Response from sensor: ");
  // for (int i = 0; i < 4; i++)
  // {
  //   Serial.print(response[i], HEX);
  //   Serial.print(" ");
  // } 
  // Serial.println();

  // Invalid response?
  if (!((response[0] == 0x16) &&
        (response[1] == 0x01) &&
        (response[2] == 0x03) &&
        (response[3] == 0xE6)))
  {
    return(false);
  }

  return(true);
}


// ----------------------------------------------------------------------------
// Private functions
// ----------------------------------------------------------------------------

/*!
 * \brief Calculate 16-bit CRC on data buffer
 * \param dataarray
 *      Buffer pointer to calculate CRC over.
  * \param datalen
 *      Buffer length.
 * \return
 *      Calculated 16-bit CRC.
 */
uint16_t C8_CO2_5K::calcuCRC(uint8_t *dataarray, uint16_t datalen)
{ 
  uint8_t uchCRCHi = 0xFF ; /* CRC High byte initialization*/ 
  uint8_t uchCRCLo = 0xFF ; /* CRC Low byte initialization*/ 
  uint16_t uIndex ; /* CRC Query table index*/ 
  uint16_t crc; const uint8_t auchCRCHi[] = 
  {
    0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 
    0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 
    0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 
    0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 
    0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 
    0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 
    0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 
    0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 
    0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 
    0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 
    0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 
    0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 
    0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 
    0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 
    0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 
    0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40
  }; 

  const uint8_t auchCRCLo[] = 
  {
    0x00, 0xC0, 0xC1, 0x01, 0xC3, 0x03, 0x02, 0xC2, 0xC6, 0x06, 0x07, 0xC7, 0x05, 0xC5, 0xC4, 0x04, 
    0xCC, 0x0C, 0x0D, 0xCD, 0x0F, 0xCF, 0xCE, 0x0E, 0x0A, 0xCA, 0xCB, 0x0B, 0xC9, 0x09, 0x08, 0xC8, 
    0xD8, 0x18, 0x19, 0xD9, 0x1B, 0xDB, 0xDA, 0x1A, 0x1E, 0xDE, 0xDF, 0x1F, 0xDD, 0x1D, 0x1C, 0xDC, 
    0x14, 0xD4, 0xD5, 0x15, 0xD7, 0x17, 0x16, 0xD6, 0xD2, 0x12, 0x13, 0xD3, 0x11, 0xD1, 0xD0, 0x10, 
    0xF0, 0x30, 0x31, 0xF1, 0x33, 0xF3, 0xF2, 0x32, 0x36, 0xF6, 0xF7, 0x37, 0xF5, 0x35, 0x34, 0xF4, 
    0x3C, 0xFC, 0xFD, 0x3D, 0xFF, 0x3F, 0x3E, 0xFE, 0xFA, 0x3A, 0x3B, 0xFB, 0x39, 0xF9, 0xF8, 0x38, 
    0x28, 0xE8, 0xE9, 0x29, 0xEB, 0x2B, 0x2A, 0xEA, 0xEE, 0x2E, 0x2F, 0xEF, 0x2D, 0xED, 0xEC, 0x2C, 
    0xE4, 0x24, 0x25, 0xE5, 0x27, 0xE7, 0xE6, 0x26, 0x22, 0xE2, 0xE3, 0x23, 0xE1, 0x21, 0x20, 0xE0, 
    0xA0, 0x60, 0x61, 0xA1, 0x63, 0xA3, 0xA2, 0x62, 0x66, 0xA6, 0xA7, 0x67, 0xA5, 0x65, 0x64, 0xA4, 
    0x6C, 0xAC, 0xAD, 0x6D, 0xAF, 0x6F, 0x6E, 0xAE, 0xAA, 0x6A, 0x6B, 0xAB, 0x69, 0xA9, 0xA8, 0x68, 
    0x78, 0xB8, 0xB9, 0x79, 0xBB, 0x7B, 0x7A, 0xBA, 0xBE, 0x7E, 0x7F, 0xBF, 0x7D, 0xBD, 0xBC, 0x7C, 
    0xB4, 0x74, 0x75, 0xB5, 0x77, 0xB7, 0xB6, 0x76, 0x72, 0xB2, 0xB3, 0x73, 0xB1, 0x71, 0x70, 0xB0, 
    0x50, 0x90, 0x91, 0x51, 0x93, 0x53, 0x52, 0x92, 0x96, 0x56, 0x57, 0x97, 0x55, 0x95, 0x94, 0x54, 
    0x9C, 0x5C, 0x5D, 0x9D, 0x5F, 0x9F, 0x9E, 0x5E, 0x5A, 0x9A, 0x9B, 0x5B, 0x99, 0x59, 0x58, 0x98, 
    0x88, 0x48, 0x49, 0x89, 0x4B, 0x8B, 0x8A, 0x4A, 0x4E, 0x8E, 0x8F, 0x4F, 0x8D, 0x4D, 0x4C, 0x8C, 
    0x44, 0x84, 0x85, 0x45, 0x87, 0x47, 0x46, 0x86, 0x82, 0x42, 0x43, 0x83, 0x41, 0x81, 0x80, 0x40
  }; 

  while (datalen--) /* Complete the entire message buffer*/
  { 
    uIndex = uchCRCLo ^ *dataarray++; /* count CRC */ 
    uchCRCLo = uchCRCHi ^ auchCRCHi[uIndex]; 
    uchCRCHi = auchCRCLo[uIndex];
  }

  crc = (uint16_t)uchCRCHi * 256; 
  crc += (uint16_t)uchCRCLo; 

  return(crc);
}